require "soar_customer/version"
require 'soar_idm/directory_provider'
require 'uri'
require 'byebug'

module SoarCustomer
  # Custom Error Handling for Provider
  class SoarHapiError < StandardError
  end

  # Directory Provider to HAPI for all things related to Customer Data
  #
  # @author Dane-Garrin Balia
  class HapiProvider < SoarIdm::DirectoryProvider
    # Passed in at initialization with connection details, username and password etc
    attr_reader :configuration
    # URL for HAPI
    attr_reader :server_url
    # Credentials is both the username and password
    attr_reader :credentials
    # Username for HAPI
    attr_reader :username
    # Password for HAPI
    attr_reader :password
    # Faraday initialized 
    attr_reader :connection

    # Constructor
    # @param configuration [Hash] contains the configuration details for connecting to HAPI
    # @note Cannot bootstrap without these details
    def initialize(configuration)
      bootstrap(configuration)
    end

    # Used to authenticate against HAPI
    # @return [Boolean] Returns true if authentication details are passed in
    # @note Must be called first prior to any use
    def authenticate(credentials)
      @credentials = credentials
      validate_credentials(credentials)
      remember_credentials(credentials)
    end

    # Bootstrapping the application via configuration
    def bootstrap(configuration)
      @configuration = configuration
      validate_configuration(@configuration)
      remember_configuration(@configuration)
    end
    
    # Used to create a profile through HAPI
    # @return [String] Returns body of message (JSON string) from HAPI response
    # @note This is how you create a customer profile (not an account)
    def create_profile(customer)
      post('/profile', customer)
    end
   
   # Confirms if the class is instantiated with some form of configuration information 
    def bootstrapped?
      not @configuration.nil?
    end

    protected
    # Generic Method for Posting to HAPI
    # @param endpoint [String] used for what resource to access on HAPI
    # @param data [JSON] customer details
    # 
    # @note Since this is a "driver" it should bubble up messages 
    def post(endpoint, data)
      raise SoarHapiError.new('Missing endpoint') if endpoint.nil? or endpoint.empty?
      raise SoarHapiError.new('Missing data') if data.nil? or data.empty?
      setup_connection
      response = submit_to_hapi(endpoint, data)
      raise SoarHapiError.new("Invalid - HTTP status: #{response.status}") if response.status != 200
      response.body
    rescue Faraday::Error::ConnectionFailed => e
      raise SoarHapiError.new('Connection failed to HAPI')
    end
    
    private

    # Submit information to HAPI (only post)
    # 
    # @note Raises exception to Method post(endpoint, data)
    def submit_to_hapi(endpoint, data)
      output = @connection.post do |req|
        req.url endpoint
        req.headers['Content-Type'] = 'application/json'
        req.body = data
      end
      output
    rescue Exception => e
      raise 
    end

    # Sets up the connection by instantiating Faraday
    # 
    # @note Raises exception to Method post(endpoint, data)
    def setup_connection
      @connection = Faraday.new(:url => @server_url)
      @connection.basic_auth(@username, @password)
    rescue Exception => e
      raise 
    end

    # Validation for configuration hash
    #
    # @note Returns a SoarHapiError if validation fails
    def validate_configuration(configuration)
      raise SoarHapiError.new('No configuration') if configuration.nil?
      raise SoarHapiError.new('Empty configuration') if configuration == {}
      raise SoarHapiError.new('Invalid configuration') if not configuration.is_a?(Hash)
      raise SoarHapiError.new('Missing server url') if configuration['server_url'].nil?
      raise SoarHapiError.new('Invalid server url') unless validate_server_url?(configuration['server_url'])
    end

    # Validation for Service Url
    def validate_server_url?(server_url)
      server_url =~ URI::regexp
    end

    # Validation for credentials 
    #
    # @note Returns a SoarHapiError if validation fails
    def validate_credentials(credentials)
      raise SoarHapiError.new('Missing credentials') if credentials.nil?
      raise SoarHapiError.new('Empty credentials') if credentials == {}
      raise SoarHapiError.new('Invalid credentials') if not credentials.is_a?(Hash)
      raise SoarHapiError.new('Missing username') if credentials['username'].nil? or credentials['username'].empty?
      raise SoarHapiError.new('Missing password') if credentials['password'].nil? or credentials['password'].empty? 
    end

    # Ensures configuration is bound for re-use
    def remember_configuration(configuration)
      @configuration = configuration
      @server_url = @configuration['server_url']
    end

    # Ensure credential is bound for re-use
    def remember_credentials(credentials)
      @credentials = credentials
      @username = @credentials['username']
      @password = @credentials['password']
      true
    end
  end
end
